const std = @import("std");

const LabjackYaesu = @import("./LabjackYaesu.zig");

const RotCtl = @This();

const log = std.log.scoped(.RotCtl);

writer: std.io.BufferedWriter(512, std.net.Stream.Writer),
running: bool,
rotator: LabjackYaesu,

pub fn run() !void {
    var server = std.net.StreamServer.init(.{ .reuse_address = true });
    defer server.deinit();

    const listen_addr = try std.net.Address.parseIp(
        config.gpredict_listen_address,
        config.gpredict_listen_port,
    );
    server.listen(listen_addr) catch {
        log.err("Could not listen on {}. Is it already in use?", .{listen_addr});
        return;
    };
    log.info("Listening for client on: {}", .{listen_addr});

    var interface: RotCtl = .{
        .writer = undefined,
        .running = true,
        .rotator = LabjackYaesu.init(),
    };

    while (true) {
        const client = try server.accept();
        defer {
            log.info("disconnecting client", .{});
            client.stream.close();
        }

        interface.writer = .{ .unbuffered_writer = client.stream.writer() };
        interface.running = true;
        defer interface.running = false;

        log.info("client connected from {}", .{client.address});

        var readbuffer = [_]u8{0} ** 512;
        var fbs = std.io.fixedBufferStream(&readbuffer);
        const reader = client.stream.reader();

        while (interface.running) : (fbs.reset()) {
            reader.streamUntilDelimiter(fbs.writer(), '\n', readbuffer.len) catch break;
            try radio.handleHamlibCommand(fbs.getWritten());
        }
    }
}

fn write(self: *const Hamlibber, buf: []const u8) !void {
    try self.writer.writeAll(buf);
}

fn replyStatus(self: *RadioProxy, comptime status: HamlibErrorCode) !void {
    try self.write(comptime status.replyFrame() ++ "\n");
}

fn handleHamlibCommand(
    self: *const Hamlibber,
    command: []const u8,
) !void {
    var tokens = std.mem.tokenizeScalar(u8, command, ' ');

    const first = tokens.next().?;
    if (first.len == 1 or first[0] == '\\') {
        switch (first[0]) {
            'F' => {
                const freqt = tokens.next() orelse
                    return self.replyStatus(.invalid_parameter);

                const new = std.fmt.parseInt(i32, freqt, 10) catch
                    return self.replyStatus(.invalid_parameter);

                self.setFrequency(new) catch
                    return self.replyStatus(.io_error);

                try self.replyStatus(.okay);
            },
            'f' => {
                const freq = self.getFrequency() catch
                    return self.replyStatus(.io_error);

                try self.print("{d}", .{freq});
            },
            't' => {
                try self.print("{d}", .{@intFromBool(self.ptt_state)});
            },
            'q' => {
                self.running = false;
                self.replyStatus(.okay) catch return;
            },
            '\\' => return self.parseLongCommand(first[1..], &tokens),
            // zig fmt: off
             '*',  '1',  '2',  '4',  '_',  'A',  'a',  'B',  'b',  'C',  'c',  'D',
             'd',  'E',  'e',  'G',  'g',  'H',  'h',  'I',  'i',  'J',  'j',  'L',
             'l',  'M',  'm',  'N',  'n',  'O',  'o',  'P',  'p',  'R',  'r',  'S',
             's',  'T',  'U',  'u',  'V',  'v',  'X',  'x',  'Y',  'y',  'Z',  'z',
            0x87, 0x88, 0x89, 0x8A, 0x8B, 0x90, 0x91, 0x92, 0x93, 0xF3, 0xF5
            => |cmd| {
                log.warn("Unsupported command {c}", .{cmd});
                try self.replyStatus(.not_supported);
            },
                // zig fmt: on
            else => |cmd| {
                log.err("unknown command {}", .{cmd});
                try self.replyStatus(.not_implemented);
            },
        }
    } else if (std.mem.eql(u8, first, "AOS")) {
        // gpredict just kind of shoves this message in on top of the HamLib
        // protocol.
        log.info("Received AOS message from gpredict", .{});
        self.setFrequency(self.base_freq) catch return self.replyStatus(.io_error);
        try self.replyStatus(.okay);
    } else if (std.mem.eql(u8, first, "LOS")) {
        log.info("Received LOS message from gpredict", .{});
        if (self.stop_on_los) {
            self.running = false;
            self.setFrequency(self.base_freq) catch return self.replyStatus(.io_error);
        }
        try self.replyStatus(.okay);
    } else try self.replyStatus(.not_supported);
}

fn parseLongCommand(
    self: *RadioProxy,
    command: []const u8,
    tokens: *std.mem.TokenIterator(u8, .scalar),
) !void {
    _ = tokens;

    for (hamlib_commands) |check| {
        if (command.len >= check.long.len and std.mem.eql(u8, check.long, command[0..check.long.len])) {
            log.warn("Unsupported command {s}", .{command});
            break;
        }
    } else {
        log.warn("Unknown command {s}", .{command});
    }
    return self.replyStatus(.not_supported);
}

fn resetWatchdog(self: *RadioProxy) !void {
    const wd = spacecraft.shared.commands.watchdog;

    const reqheader = csp.Header{
        .priority = 1,
        .source = Config.global.doppler_shift.gs_source_address,
        .destination = Config.global.doppler_shift.gs_radio_address,
        .destination_port = @intFromEnum(wd.Reset.port),
        .source_port = self.reqPort(),
    };

    const request = wd.Reset{};
    const packet = reqheader.packBig() ++ request.packLittle();

    try self.nats.publish(self.nats_up, &packet);

    while (true) {
        const res = self.nats_down.nextMessage(reply_timeout) catch |err| {
            log.err("Could not reset ground UHF radio watchdog. Is the radio on and bridged to NATS?", .{});
            return err;
        };
        const data = res.getData() orelse continue;
        if (data.len != csp.Header.pack_size + wd.Reset.Response.pack_size)
            continue;

        const resheader = csp.Header.unpackBig(data[0..csp.Header.pack_size].*);
        if (!resheader.isReply(reqheader)) continue;

        const reply = wd.Reset.Response.unpackLittle(
            data[csp.Header.pack_size..][0..wd.Reset.Response.pack_size].*,
        );
        if (reply.err != .okay) {
            log.err("Got error response for ground UHF watchdog reset: {}", .{reply.err.code()});
            return error.Failure;
        }
        break;
    }

    log.info("Successfully reset the ground UHF radio watchdog timer.", .{});
}

const HamlibErrorCode = enum(u8) {
    okay = 0,
    invalid_parameter = 1,
    invalid_configuration = 2,
    out_of_memory = 3,
    not_implemented = 4,
    timeout = 5,
    io_error = 6,
    internal_error = 7,
    protocol_error = 8,
    command_rejected = 9,
    parameter_truncated = 10,
    not_supported = 11,
    not_targetable = 12,
    bus_error = 13,
    bus_busy = 14,
    invalid_arg = 15,
    invalid_vfo = 16,
    domain_error = 17,
    deprecated = 18,
    security = 19,
    power = 20,

    fn replyFrame(comptime self: HamlibErrorCode) []const u8 {
        return std.fmt.comptimePrint(
            "RPRT {d}",
            .{-@as(i8, @intCast(@intFromEnum(self)))},
        );
    }
};

const HamlibCommand = struct {
    short: ?u8 = null,
    long: ?[]const u8 = null,
    mode: enum { rotator, radio, both },
};

const hamlib_commands = [_]HamlibCommand{
    .{ .short = 'F', .long = "set_freq", .mode = .radio },
    .{ .short = 'f', .long = "get_freq", .mode = .radio },
    .{ .short = 'T', .long = "set_ptt", .mode = .radio },
    .{ .short = 't', .long = "get_ptt", .mode = .radio },
    .{ .short = 'q', .mode = .both }, // quit
    .{ .short = 'Q', .mode = .both }, // quit
    .{ .short = 'P', .long = "set_pos", .mode = .rotator }, // azimuth: f64, elevation: f64
    .{ .short = 'p', .long = "get_pos", .mode = .rotator }, // return az: f64, el: f64
    .{ .short = 'M', .long = "move", .mode = .rotator }, // direction: enum { up=2, down=4, left=8, right=16 }, speed: i8 (0-100 or -1)
    .{ .short = 'S', .long = "stop", .mode = .rotator },
    .{ .short = 'K', .long = "park", .mode = .rotator },
    .{ .short = 'C', .long = "set_conf", .mode = .rotator }, // token: []const u8, value: []const u8
    .{ .short = 'R', .long = "reset", .mode = .rotator }, // u1 (1 is reset all)
    .{ .short = '_', .long = "get_info", .mode = .rotator }, // return Model name
    .{ .short = 'K', .long = "park", .mode = .rotator },
    .{ .long = "dump_state", .mode = .rotator }, // ???
    .{ .short = '1', .long = "dump_caps", .mode = .rotator }, // ???
    .{ .short = 'w', .long = "send_cmd", .mode = .rotator }, // []const u8, send serial command directly to the rotator
    .{ .short = 'L', .long = "lonlat2loc", .mode = .rotator }, // return Maidenhead locator for given long: f64 and lat: f64, locator precision: u4 (2-12)
    .{ .short = 'l', .long = "loc2lonlat", .mode = .rotator }, //  the inverse of the above
    .{ .short = 'D', .long = "dms2dec", .mode = .rotator }, // deg, min, sec, 0 (positive) or 1 (negative)
    .{ .short = 'd', .long = "dec2dms", .mode = .rotator },
    .{ .short = 'E', .long = "dmmm2dec", .mode = .rotator },
    .{ .short = 'e', .long = "dec2dmmm", .mode = .rotator },
    .{ .short = 'B', .long = "grb", .mode = .rotator },
    .{ .short = 'A', .long = "a_sp2a_lp", .mode = .rotator },
    .{ .short = 'a', .long = "d_sp2d_lp", .mode = .rotator },
    .{ .long = "pause", .mode = .rotator },
};

// D, dms2dec 'Degrees' 'Minutes' 'Seconds' 'S/W'
// Returns 'Dec Degrees', a signed floating point value.
// 'Degrees' and 'Minutes' are integer values.
// 'Seconds' is a floating point value.
// 'S/W' is a flag with ’1’ indicating South latitude or West longitude and ’0’ North or East (the flag is needed as computers don’t recognize a signed zero even though only the 'Degrees' value is typically signed in DMS notation).
// d, dec2dms 'Dec Degrees'
// Returns 'Degrees' 'Minutes' 'Seconds' 'S/W'.
// Values are as in dms2dec above.
// E, dmmm2dec 'Degrees' 'Dec Minutes' 'S/W'
// Returns 'Dec Degrees', a signed floating point value.
// 'Degrees' is an integer value.
// 'Dec Minutes' is a floating point value.
// 'S/W' is a flag as in dms2dec above.
// e, dec2dmmm 'Dec Deg'
// Returns 'Degrees' 'Minutes' 'S/W'.
// Values are as in dmmm2dec above.
// B, qrb 'Lon 1' 'Lat 1' 'Lon 2' 'Lat 2'
// Returns 'Distance' and 'Azimuth'.
// 'Distance' is in km.
// 'Azimuth' is in degrees.
// Supplied Lon/Lat values are signed floating point numbers.
// A, a_sp2a_lp 'Short Path Deg'
// Returns 'Long Path Deg'.
// Both the supplied argument and returned value are floating point values within the range of 0.00 to 360.00.
// Note: Supplying a negative value will return an error message.
// a, d_sp2d_lp 'Short Path km'
// Returns 'Long Path km'.
// Both the supplied argument and returned value are floating point values.
// pause 'Seconds'
// Pause for the given whole (integer) number of 'Seconds' before sending the next command to the rotator.
